/**
 * @file  AE3Channel.cpp
 * @brief Implementierung der Klasse CAE3Channel.
 */
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "AE3Channel.h"

#include "../Common/eeprom.h"
#include "../Common/CRC.h"

#include <time.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CAE3Channel::CAE3Channel(unsigned char ucNum) : CAE3Buffer(ucNum), CAE3Modbus(ucNum)
{
	for (int i=0; i < MAX_MB_PAKET; i++) {
	
		m_pucRecordBuf[i] = 0;
	}

	m_iRecordLength = 0;
	m_iRecordError = 0;

	m_usLastCollectVersion = 0;
	m_usLastDeviceVersion = 0;

	m_iDevicePollRetry = 0;
}

CAE3Channel::~CAE3Channel()
{

}

void CAE3Channel::ClearMem()
{
	// Speicher initialisieren
	memset(m_pucEepMem, 0xFF, sizeof(m_pucEepMem));
	memset(m_pucFlashMem, 0xFF, sizeof(m_pucFlashMem));
	memset(m_pucConfigMem, 0x00, sizeof(m_pucConfigMem));
	memset(m_pucExtendedMem, 0x00, sizeof(m_pucExtendedMem));
}

void CAE3Channel::ResetWithError(int err) 
{
	if (err == ERROR_CRC || err == ERROR_COM || 
		err == ERROR_NO_ANSWER || err == ERROR_AT_RECIEVING) {
	
		ClearCom();
	}

	TRACE("CAE3Channel::ResetWithError: (%i)\n", err);
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

// Nach Wechsel der Busnummer ist Datenpuffer ungueltig
int CAE3Channel::ChangeBusNum(unsigned char ucNewBusNum)
{
	for (int i=0; i < MAX_MB_PAKET; i++) {
	
		m_pucTmpRecordBuf[i] = 0;
		m_pucRecordBuf[i] = 0;
	}

	m_iRecordLength = 0;
	m_iRecordError = 0;

	return CAE3Modbus::ChangeBusNum(ucNewBusNum);
}

// Lesen eines Datensatzes.
// Rueckgabewerte: -1 - Lesen gescheitert
//					0 - Keine Datensaetzte vorhanden
//					1 - Online Datensatz gelesen
//					2 - Wiederholt Online Datensatz gelesen
//					3 - Offline Datensatz gelesen
//					4 - Wiederholt Offline Datensatz gelesen
int CAE3Channel::ReadRecord(unsigned char ucBusNum, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
	if (m_ChannelType == Undef) {
		
		*piError = ERROR_COM_CLOSED;
		return -1;
	}

	int iResult = -1;
	//Senden des Lese-Datensatz-Befehl
	m_pucTxBuf[0] = ucBusNum;			// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x06;				// Befehlslaenge = 6
	m_pucTxBuf[3] = MB_CMD_READ_RECORD;	// Befehl: Lesen eines Datensatzes
	MkMBPaket(m_pucTxBuf, 4);
	
	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 3) && !bEnd; i++) { // 3 Versuche 

		if(!Command(m_pucTxBuf, 6, pucRxBuf, &iResult, piError)) {
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();

		} else {
		
			bEnd = true;
		}
	}

	// Fehlerhafte versuche
	if (!bEnd) {

		return -1;
	}

	// Paket kleiner als Mindestlaenge?
	if(iResult < 6){

		*piError = ERROR_AT_RECIEVING;
		return -1;
	}
	
	// Pruefsumme OK?
	if(!CheckMBPaket(pucRxBuf, iResult)){
		
		if (!bCRCFail) { // nach CRC - Fehler nochmals versuchen

			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return -1;
	}
	
	*piRxCount = iResult;

	// Pruefen der Antwort fuer Rueckgabe
	switch(pucRxBuf[3]) {
	case MB_ANSW_NO_RECORD:         // kein Datensatz vorhanden 

		return 0;

	case MB_ANSW_ONLINE_RECORD:     // Online Datensatz 

		return 1;

	case MB_ANSW_OLD_ONLINE_RECORD: // wiederholt gesendeter Online Datensatz 

		return 2;

	case MB_ANSW_OFFLINE_RECORD:    // Offline Datensatz 

		return 3;

	case MB_ANSW_OLD_OFFLINE_RECORD:// wiederholt gesendeter Offline Datensatz 

		return 4;
	};

	// Keine gueltige Antwort auf Datensatzanfrage
	*piError = ERROR_SENDER;

	return -1;
}

// Quittieren eines Datensatzes
// Rueckgabewerte:	-1 - Fehler bei Quittierung
//					 0 - Keine Quittung erforderlich
//					 1 - Quittierung erfolgt (online)
//					 2 - Quittierung erfolgt (wiederholte online)
//					 3 - Quittierung erfolgt (offline)
//					 4 - Quittierung erfolgt (wiederholte offline)
int CAE3Channel::QuitRecord(unsigned char ucBusNum, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
	int iResult = FALSE;
	
	if (m_ChannelType == Undef){
		
		*piError = ERROR_COM_CLOSED;
		return -1;
	}

	// Senden des Lese - Befehls
	m_pucTxBuf[0] = ucBusNum;			// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x06;				// Befehlslaenge = 6
	m_pucTxBuf[3] = MB_CMD_ACK_RECORD;	// Befehl: Letzen Datensatz quittieren
	MkMBPaket(m_pucTxBuf, 4);
	
	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, 6, pucRxBuf, &iResult, piError)) {
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();

		} else {
			
			bEnd = true;
		}
	}

	// Fehlerhafte Anfrage
	if (!bEnd) {

		return -1;
	}

	// Paket kleiner als Mindestlaenge?
	if(iResult < 6) {
		
		*piError = ERROR_AT_RECIEVING;
		return -1;
	}

	// Pruefsumme OK?
	if(!CheckMBPaket(pucRxBuf, iResult)) {
		
		if (!bCRCFail) { // Nach CRC - Fehler nochmals anfordern
			
			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return -1;
	}
	
	*piRxCount = iResult;

	switch (pucRxBuf[3]) {
	case 0: // Keine vorherige Datenabholung
	case 2: // Keine vorherige Datenabholung (wiederholt)
		return 0; // Keine Quitierung erforderlich
		 
	case MB_ANSW_ACK_ON_RECORD:          // Quittung erhalten (fuer Online-Datensatz)
	case MB_ANSW_TIMEOUT_ACK_RECORD:     // Quittung kam nach Timeout (fuer Online-Datensatz)
		return 1;

	case MB_ANSW_OLD_ACK_ON_RECORD:      // wiederholt gesendete Quittung
	case MB_ANSW_OLD_TIMEOUT_ACK_RECORD: // wiederholte Quittung nach Timeout
		return 2;

	case MB_ANSW_ACK_OFF_RECORD:	     // Quittung erhalten (fuer Offline-Datensatz)
		return 3;

	case MB_ANSW_OLD_ACK_OFF_RECORD:     // wiederholt gesendete Quittung
		return 4;
	}

	// Keine gueltige Antwort auf Datensatzanfrage
	*piError = ERROR_SENDER;
	return -1;
}

int CAE3Channel::DFReadRecord(int deviceID, char *buf, int *length, int *error)
{
	int iRes;
	
	// Wechsel der Busnummer und somit loeschen des Datenpuffer
	if (!ChangeBusNum(deviceID)) {
	
		*error = ERROR_BUS_NUM;
		return -1;
	}
	
	// Pruefen der Kommunikationsversionsnummer
	if (!(iRes = CheckComVersion(GetBusNum(), error))) {
	
		//*piError = ERROR_FW_VERSION;
		return -1;
	}

	// Fuer weitere Pruefung halten
	m_usLastCollectVersion = iRes;

	switch (iRes) {
	case 0x0200: // AEIII+
	case 0x0201:
	case 0x0300: 
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZE - Master IV
	case 0x0501:
	case 0x0502:

		// Datensatz ermitteln
		iRes = ReadRecord(GetBusNum(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);

		switch ( iRes ) {
		case 0: // Keine Datensaetze vorhanden
			
			memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
			m_iRecordLength = 0;
			m_iRecordError = 0;

			return 0;

		case 1: // Online Datensatz
		case 2: // Wiederholter Online Datensatz
		case 3: // Offline Datensatz
		case 4: // Wiederholter Offline Datensatz
			{
			// Daten in Puffer uebergeben
			int i;
			// Zu groe Datensaetze abfangen!
			if ((m_iRecordLength - 6) > MAX_MB_PAKET_DATA) {
				m_iRecordLength -= (m_iRecordLength - 6) - MAX_MB_PAKET_DATA;
			}
			for (i=0; i < (m_iRecordLength - 6); i++) {
			
				buf[i] = m_pucRecordBuf[i + 4];
			}

			*length = m_iRecordLength - 6;

			if ((m_usLastCollectVersion == 0x0200) &&
				((m_pucRecordBuf[3] == MB_ANSW_OFFLINE_RECORD) || 
				 (m_pucRecordBuf[3] == MB_ANSW_OLD_OFFLINE_RECORD)) ) {

				// Fehler bei Offline in CW 0x200: konstant 3 zuviel
				// Fehler bei Offline in CW 0x201 bei FW 2.8: konstant 3 zuviel aber 2.9 nicht!
				*length -= 3; 
			}

			return iRes;}
		}

		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ClearCom();
		break;

	default:
		*error = ERROR_FW_VERSION;
	}

	return -1;
}

int CAE3Channel::DFQuitRecord(int deviceID, int *error)
{
	int iRes;
	
	// Wechsel der Busnummer und somit loeschen des Datenpuffer
	if (!ChangeBusNum(deviceID)) {
	
		*error = ERROR_BUS_NUM;
		return -1;
	}
	
	// Pruefen der Kommunikationsversionsnummer
	if (!(iRes = CheckComVersion(GetBusNum(), error))) {
	
		//*piError = ERROR_FW_VERSION;
		return -1;
	}

	switch (iRes) {
	case 0x0200: // AEIII+
	case 0x0201:
	case 0x0300: 
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZE - Master IV
	case 0x0501:
	case 0x0502:

		// Datensatz quitieren
		iRes = QuitRecord(GetBusNum(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		
		switch ( iRes ) {
		case 0: // Keine Datensaetze vorhanden
			
			return 0;

		case 1: // Online Datensatz
		case 2: // Wiederholter Online Datensatz
		case 3: // Offline Datensatz
		case 4: // Wiederholter Offline Datensatz

			return iRes;
		}

		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ClearCom();
		break;

	default:
		*error = ERROR_FW_VERSION;
	}

	return -1;
}

// Loescht alle Datensaetze
int CAE3Channel::ClearData(unsigned char ucBusNum, int *piError)
{
	if (m_ChannelType == Undef){

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int iResult = FALSE;

	// Senden des Lesen-Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x06;		// Befehlslaenge = 6
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:

		m_pucTxBuf[3] = CMD_DEL;	// Alter Befehl: Loeschen
		break;

	case 0x0200:
	case 0x0201:
	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		m_pucTxBuf[3] = MB_CMD_DEL;	// Neuer Befehl: Loeschen
		break;

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}


	MkMBPaket(m_pucTxBuf, 4);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, 6, m_pucRxBuf, &iResult, piError)){
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != 6){
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// Antwort muss genauso aussehen wie der Befehl, also jedes Byte ueberpruefen
	// Checksumme muss daher nicht ueberprueft werden
	for(int counter=0; counter < iResult; counter++) {
		
		if(m_pucTxBuf[counter] != m_pucRxBuf[counter]){

			if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

				// Nachfolgende Zeichen verwerfen
				ClearCom();

				bCRCFail = true;
				goto crc_fail;
			}

			*piError = ERROR_CRC;
			return FALSE;
		}
	}

	return TRUE;
}

int CAE3Channel::GetDateTime(unsigned char ucBusNum, unsigned char *pucDateTime, int *piError)
{
	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}


	if(!ReadMem(MB_GET_TIME, 7, pucDateTime, piError)) {
			
		return FALSE;
	}

	return TRUE;
}

int CAE3Channel::SetDateTime(unsigned char ucBusNum, unsigned char *pucDateTime, int *piError)
{
	unsigned char ucDateTime[8];

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	if (pucDateTime == NULL) {
	
		time_t tm_time;
		struct tm *ptm;
		tm_time = time(NULL);
		ptm = localtime(&tm_time);
		
		ucDateTime[0] = (ptm->tm_year + 1900) / 100;
		ucDateTime[1] = (ptm->tm_year + 1900) % 100;
		ucDateTime[2] = ptm->tm_mon + 1;
		ucDateTime[3] = ptm->tm_mday;
		ucDateTime[4] = ptm->tm_hour;
		ucDateTime[5] = ptm->tm_min;
		ucDateTime[6] = ptm->tm_sec;

		if(!WriteMem(MB_SET_TIME, 7, ucDateTime, piError)) {
				
			return FALSE;
		}
	} else {

		if(!WriteMem(MB_SET_TIME, 7, pucDateTime, piError)) {
				
			return FALSE;
		}
	}

	return TRUE;

}

int CAE3Channel::SendInfotext(unsigned char ucBusNum, char *szMessage, int iLength, int *piError)
{
	unsigned char Buffer[245];
	int counter;

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	Buffer[0] = 0xFF;
	Buffer[1] = 0;
	Buffer[2] = 0;

	for(counter=0; counter < iLength; counter++) {

		Buffer[counter + 3] = szMessage[counter];
	}

	if(!WriteMem(INFOTEXT_MEMORY, iLength + 3, &Buffer[0], piError)) {
		
		return FALSE;
	}

	return TRUE;
}

int CAE3Channel::SendMessage(unsigned char ucBusNum, unsigned char ucDelayTime, 
							 unsigned char ucAction,  unsigned char ucAudio, 
							 char *szMessage, int iLength, int *piError)
{
	unsigned char Buffer[245];
	int counter;

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	Buffer[0] = ucDelayTime;
	Buffer[1] = ucAction;
	Buffer[2] = ucAudio;

	for(counter=0; counter < iLength; counter++) {
		
		Buffer[counter + 3] = szMessage[counter];
	}

	if(!WriteMem(DISPLAY_MEMORY, iLength + 3, &Buffer[0], piError)) {
		
		return FALSE;
	}

	return TRUE;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int CAE3Channel::GetTableDescription(unsigned char ucBusNum, unsigned char ucList, int *piError)
{
	ULONG pDat, pFeld, pList[MAX_DATENSATZ_BESCHREIBUNG], pListEnd;
	int i, j;

	// FIX: Wenn Daten zuvor mit CollectData abgeholt wurden und danach fuer die
	// selbe BusNum die Datensatzbeschreibung geladen wird, dann wird der zuvor
	// geladene Datensatz gehalten
	if (!CAE3Modbus::ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// EEPROM lesen 
 	if (!ReadMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError)) {

		return FALSE;
	}

	// Start und Ende der Datensatzbeschreibungen lesen
	ChangeOrderLong(&m_pucEepMem[P_DAT_STRUCT]);
	StartDatensatzB = *(ULONG*)(&m_pucEepMem[P_DAT_STRUCT]);

	ChangeOrderLong(&m_pucEepMem[P_LIST_STRUCT]);
	EndDatensatzB = *(ULONG*)(&m_pucEepMem[P_LIST_STRUCT]);

	if (StartDatensatzB >=  MAX_FLASH_MEM || EndDatensatzB >= MAX_FLASH_MEM) {
		
		*piError = ERROR_COM;
		return FALSE;
	}

	if (ucList) {

		//Pointer der Listenbeschreibungen lesen (Ende Datensatzbeschreibung ist anfang Listenb.
		if (!ReadMem(EndDatensatzB, MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG), &m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START], piError)) {
		
			return FALSE;
		}

		// Pointer extrahieren und Anzahl Listenb. ermitteln
		for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

			ChangeOrderLong(&m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
			pList[i] = *(ULONG*)(&m_pucFlashMem[EndDatensatzB - FLASH_KONFIG_START + i *sizeof(ULONG)]);

			if (pList[i] == 0ul) {
				
				MaxListB = i; 
				i = MAX_DATENSATZ_BESCHREIBUNG; 
				continue;
			}
		}

		// Ende Listenbereich angenaehert berechnen (mit Reserve)
		pListEnd = pList[MaxListB - 1] + sizeof(DatensatzBeschreibung[0]) + 
					(sizeof(DatensatzFeld[0][0]) + 
					sizeof(ULONG)) * (MAX_DATENSATZ_BESCHREIBUNG + 1);

		// Rest der Listenbeschreibung aus AEIII+ lesen
		pDat = EndDatensatzB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG);

		if (!ReadMem(pDat, pListEnd - pDat, &m_pucFlashMem[pDat - FLASH_KONFIG_START], piError)) {
			
			return FALSE;
		}

		// Listenbeschreibungen auswerten
		for (i=0; i < MaxListB; i++) {

			CpyFromKonfigBuf((unsigned char*)&ListenBeschreibung[i], m_pucFlashMem, pList[i], sizeof(ListenBeschreibung[i]));
			for (j=0; j < ListenBeschreibung[i].Felder; j++) { // Feldbeschreibung einlesen
				
				ChangeOrderLong  (&m_pucFlashMem[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
				pFeld = *(ULONG*)(&m_pucFlashMem[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

				ASSERT (pFeld + sizeof(DatensatzFeld[i][j]) < pListEnd);

				CpyFromKonfigBuf((unsigned char*)&ListenFeld[i][j], m_pucFlashMem, pFeld, sizeof(ListenFeld[i][j]));
			}
		}
	} else {

		// Datensatzbeschreibungen lesen
		if (!ReadMem(StartDatensatzB, EndDatensatzB - StartDatensatzB, &m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START], piError)) {
			
			return FALSE;
		}

		// Datensatzbeschreibung auswerten
		for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++) {

			ChangeOrderLong(&m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
			pDat = *(ULONG*)(&m_pucFlashMem[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);

			if (pDat == 0ul) {
				
				MaxDatB = i; 
				i = MAX_DATENSATZ_BESCHREIBUNG; 
				continue;
			}

			CpyFromKonfigBuf((unsigned char*)&DatensatzBeschreibung[i], m_pucFlashMem, pDat, sizeof(DatensatzBeschreibung[i]));
			
			// Feldbeschreibung einlesen
			for (j=0; j < DatensatzBeschreibung[i].Felder; j++) { 
			
				ChangeOrderLong(&m_pucFlashMem[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
				pFeld = *(ULONG*)(&m_pucFlashMem[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

				CpyFromKonfigBuf((unsigned char*)&DatensatzFeld[i][j], m_pucFlashMem, pFeld, sizeof(DatensatzFeld[i][j]));
			}
		}
	}

	return TRUE;
}

// Listenbeschreibung aus dem Geraet lesen.
int CAE3Channel::GetListName(unsigned char ucBusNum, unsigned char ucList, char *name, int *piError)
{
	ULONG add;
	TDatensatzBeschreibung dsb;

	// FIX: Wenn Daten zuvor mit CollectData abgeholt wurden und danach fuer die
	// selbe BusNum die Datensatzbeschreibung geladen wird, dann wird der zuvor
	// geladene Datensatz gehalten
	if (!CAE3Modbus::ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Die Addresse auf die Listenbeschreibungen aus dem EEPROM-Bereich lesen 
 	if (!ReadMem(MB_START_EEP + P_LIST_STRUCT, sizeof(ULONG), (UCHAR*)&add, piError)) {

		return FALSE;
	}
	ChangeOrderLong((UCHAR*)&add);

	// Die Addresse auf die Listenbeschreibung lesen.
	if (!ReadMem(add + ucList * sizeof(ULONG), sizeof(ULONG), (UCHAR*)&add, piError)) {
	
		return FALSE;
	}
	ChangeOrderLong((UCHAR*)&add);


	// Wenn die Rueckgabe FALSE und Error = ERROR_NO, dann war die Liste nicht vorhanden.
	*piError = ERROR_NO;

	if (add) {

		// Die Listenbeschreibung lesen.
		if (!ReadMem(add, sizeof(TDatensatzBeschreibung), (UCHAR*)&dsb, piError)) {
		
			return FALSE;
		}

		strcpy(name, (char*)dsb.Name);
		return TRUE;
	}

	return FALSE;
}

int CAE3Channel::UpLoadSetupData(unsigned char ucBusNum, int *piError)
{

	unsigned char ucIsData = 0, temp = 1; // fuer Bedienstopp

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Abfrage ob Daten vorhanden. 
	// (wegen alter FW, nicht mit CheckData ausgefuehrt.)
	if (!ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {
		
		return FALSE;
	}
	// Abbruch wenn Daten im Geraet vorliegen
	if (ucIsData) {

		*piError = ERROR_IS_DATA;
		return FALSE;
	}

	// Abhaengig von der Kommunikationsversion sind entweder
	// 8, 12 oder 16kByte Setupdaten moeglich.
	ULONG ulSize, ulSetupSize = GetFlashSetupSize();
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:

		// Hinteren 8kByte mit 0 Initialisieren
		//memset(&m_pucFlashMem[0x2000], 0, 0x2000);
		ulSize = FLASH_SIZE_SU - 0x2000;
		break;

	case 0x0200:
	case 0x0201:

		// Hinteren 4kByte mit 0 Initialisieren
		//memset(&m_pucFlashMem[0x3000], 0, 0x1000);
		ulSize = FLASH_SIZE_SU - 0x1000;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		ulSize = FLASH_SIZE_SU;
		break;

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Vorhandene zur moeglichen Groesse pruefen
	if (ulSetupSize > ulSize) {

		*piError = ERROR_OPEN;
		return FALSE;
	}

	// Bedienstopp setzen
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError, 32)) {

		return FALSE;
	}

	// Groesse der Flashdaten in EepMem eintragen
	*(ULONG*)&m_pucEepMem[P_FLASH_SIZE_SU] = ulSetupSize;
	ChangeOrderLong(&m_pucEepMem[P_FLASH_SIZE_SU]);

	// Groesse der MDE - Eingabeketten in EepMem loeschen.
	*(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE] = 0xFFFFFFFF;
	ChangeOrderLong(&m_pucEepMem[P_MDE_EK_SIZE]);
	
	// Groesse der Zutrittskontroll - Eingabeketten in EepMem loeschen.
	*(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE] = 0xFFFFFFFF;
	ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);

	if ((m_usLastDeviceVersion == 0x0400) || 
		(m_usLastDeviceVersion == 0x0401) ||
		(m_usLastDeviceVersion == 0x0402) ||
		(m_usLastDeviceVersion == 0x0500) ||
		(m_usLastDeviceVersion == 0x0501) ||
		(m_usLastDeviceVersion == 0x0502) ) {

		// Groesse der MDE - Eingabeketten in EepMem eintragen.
		*(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
		ChangeOrderLong(&m_pucEepMem[P_MDE_EK_SIZE]);
	
		if ((m_usLastDeviceVersion == 0x0401) ||
			(m_usLastDeviceVersion == 0x0402) ||
			(m_usLastDeviceVersion == 0x0500) ||
			(m_usLastDeviceVersion == 0x0501) ||
			(m_usLastDeviceVersion == 0x0502) ) {

			// Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
			*(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE] = GetEntranceEKFlashSize();
			ChangeOrderLong(&m_pucEepMem[P_ENTRANCE_EK_SIZE]);
		}
	}

#define MB_DATABLOCKSIZE 32

	// EEPROM - schreiben 
 	if (!WriteMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError, MB_DATABLOCKSIZE)) {

		return FALSE;
	}

	// Flash - schreiben
	if (!WriteMem(FLASH_KONFIG_START, ulSetupSize, m_pucFlashMem, piError, MB_DATABLOCKSIZE)) {

		return FALSE;
	}

	// Konfigurationsdaten - schreiben
	if (!WriteMem(MB_START_CONFIG + 2, CONFIG_SIZE_SU - 2, &m_pucConfigMem[2], piError, MB_DATABLOCKSIZE)) {

		return FALSE;
	}

	// IOModule und MDE ab Version 3.0
	if ((m_usLastDeviceVersion == 0x0400) || 
		(m_usLastDeviceVersion == 0x0401) ||
		(m_usLastDeviceVersion == 0x0402) ||
		(m_usLastDeviceVersion == 0x0500) ||
		(m_usLastDeviceVersion == 0x0501) ||
		(m_usLastDeviceVersion == 0x0502) ) {

		int i;
		TAE3Extended extBegin;
		for (i=0; i < sizeof(TAE3Extended); i++) {
		
			((UCHAR*)&extBegin)[i] = m_pucExtendedMem[i];
		}


		// IOModul - Konfiguration - schreiben
		if (!WriteMem(MB_START_IOMODULE, sizeof(TAE3IOModule), &m_pucExtendedMem[extBegin.pulIOModule], piError, MB_DATABLOCKSIZE)) {

			return FALSE;
		}

		// IOModul - DigitalMDE - schreiben
		for (i=0; i < 8; i++) {
			if (!WriteMem(MB_DIGITAL_IO + i * sizeof(TAE3DigitalMDE), sizeof(TAE3DigitalMDE), 
				&m_pucExtendedMem[extBegin.pulIODigital + i * sizeof(TAE3DigitalMDE)], piError, MB_DATABLOCKSIZE)) {

				return FALSE;
			}
		}

		// Interne MDE - DigitalMDE - schreiben
		for (i=0; i < 4; i++) {
			if (!WriteMem(MB_DIGITAL_MDE + i * sizeof(TAE3DigitalMDE), sizeof(TAE3DigitalMDE), 
				&m_pucExtendedMem[extBegin.pulMDEDigital + i * sizeof(TAE3DigitalMDE)], piError, MB_DATABLOCKSIZE)) {

				return FALSE;
			}
		}

		// MDE - Eingabeketten - schreiben
		// Bei 4.0 kommt erst 1Byte Anzahl im Array
		// Ab 4.1 kommen erst 4Byte Anzahl im Array
		ULONG ulSize = 0;
		if (m_usLastDeviceVersion == 0x0400) {

			ulSize = GetMDEEKFlashSize_0400();
		} else {
			
			ulSize = GetMDEEKFlashSize();
		}

		if (!WriteMem(MB_MDE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArray], piError, MB_DATABLOCKSIZE)) {

			return FALSE;
		}

		// Zutrittskontrolle
		if ((m_usLastDeviceVersion == 0x0401) ||
			(m_usLastDeviceVersion == 0x0402) ||
			(m_usLastDeviceVersion == 0x0500) ||
			(m_usLastDeviceVersion == 0x0501) ||
			(m_usLastDeviceVersion == 0x0502) ) {

			// Zutrittskontroll - Eingabeketten - schreiben
			if (!WriteMem(MB_ENTRANCE_EK, GetEntranceEKFlashSize(), 
				&m_pucExtendedMem[extBegin.pulEKArrayEntrance], piError, MB_DATABLOCKSIZE)) {

				return FALSE;
			}
		}
	}


	temp = 0; // Bedienstopp ruecksetzen
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError, 32)) {

		return FALSE;
	}

	// Aktuelle Setup - Daten wurden erst durch zuruecksetzen von Bedienstopp uebernommen!

	// Standardwerte fuer Globale Variable
	if ((m_usLastDeviceVersion == 0x0401) ||
		(m_usLastDeviceVersion == 0x0402) ||
		(m_usLastDeviceVersion == 0x0500) ||
		(m_usLastDeviceVersion == 0x0501) ||
		(m_usLastDeviceVersion == 0x0502) ) {

		int i;
		TAE3Extended extBegin;
		for (i=0; i < sizeof(TAE3Extended); i++) {
		
			((UCHAR*)&extBegin)[i] = m_pucExtendedMem[i];
		}

		// GV setzen?
		if (extBegin.pulValueGVx) {
		
			UCHAR ucNum;
			// Struktur lesen
			TValueGVx ValueGVx;
			for (i=0; i < sizeof(TValueGVx); i++) {

				((UCHAR*)&ValueGVx)[i] = m_pucExtendedMem[extBegin.pulValueGVx + i];
			}

			// Zeit fuer Umschaltung aus Bedienstopp
			Sleep(200);

			for (i=0; i < 8; i++) {

				if (ValueGVx.ucFlags & (1 << i)) {
				
					ucNum = 1 + i;
					if (!SetGlobVar(ucBusNum, &ucNum, 0, ValueGVx.szValue[i], piError)) {
					
						return FALSE;
					}
				}
			}
		}	
	}

	// Standardwerte fuer Systemvariable
	// Ab 04.00.23
	if ((m_usLastDeviceVersion == 0x0502) ) {

		//############################
		// 2005-09-19: Erweiterte Struktur fuer Parameter
		if (((m_pucEepMem[P_SETUP_VERSION    ] == 3) &&
			 (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
			 ((m_pucEepMem[P_SETUP_VERSION + 2] > 34) ||
			  (m_pucEepMem[P_SETUP_VERSION + 2] == 34) &&
			  (m_pucEepMem[P_SETUP_VERSION + 3] >  5)) ) ||

			 ((m_pucEepMem[P_SETUP_VERSION   ] == 4) &&
			  (m_pucEepMem[P_SETUP_VERSION + 1] == 1) &&
			  ((m_pucEepMem[P_SETUP_VERSION + 2] > 0) || 
			   (m_pucEepMem[P_SETUP_VERSION + 2] == 0) &&
			   (m_pucEepMem[P_SETUP_VERSION + 3] > 2))) ) {

			int i;
			TAE3Extended extBegin;
			for (i=0; i < sizeof(TAE3Extended); i++) {
			
				((UCHAR*)&extBegin)[i] = m_pucExtendedMem[i];
			}

			// Liegt die Struktur vor?
			if (extBegin.pulExtended) {

				TExtended extParam;
				for (i=0; i < sizeof(TExtended); i++) {

					((UCHAR*)&extParam)[i] = m_pucExtendedMem[extBegin.pulExtended + i];
				}

				UINT8 value[32], name[64];
				// Buchungswiederholsperrdauer
				sprintf((char*)name, "_SERVICE.DBLINPUT");
				sprintf((char*)value, "%i", extParam.doubleInputTime);
				// 
				if (!SetGlobVar(ucBusNum, name, 2, value, piError)) {
				
					if (// Fehler liegt bei Wert
						*piError != ERROR_VAR_VALUE &&
						// Variable nicht bekannt
						*piError != ERROR_NOT_DEFINED)

					return FALSE;
				}
				// Abblendedauer fuer Bestaetigungsfeld
				sprintf((char*)name, "_SERVICE.DEMANDTIMER");
				sprintf((char*)value, "%i", extParam.confirmationTime);
				// 
				if (!SetGlobVar(ucBusNum, name, 2, value, piError)) {
				
					if (// Fehler liegt bei Wert
						*piError != ERROR_VAR_VALUE &&
						// Variable nicht bekannt
						*piError != ERROR_NOT_DEFINED)

					return FALSE;
				}
			}
		}
	}

	return TRUE;
}

int CAE3Channel::DownLoadSetupData(unsigned char ucBusNum, int *piError)
{

	unsigned char temp = 1; // fuer Bedienstopp

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}


	unsigned long ulTmp, ulSize;
	// Abhaengig von der Kommunikationsversion werden entweder
	// 8, 12 oder 16kByte Setupdaten gelesen
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:

		ulSize = FLASH_SIZE_SU - 0x2000;
		break;

	case 0x0200:
	case 0x0201:

		ulSize = FLASH_SIZE_SU - 0x1000;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		ulSize = FLASH_SIZE_SU;
		break;

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Bedienstopp setzen.
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

		return FALSE;
	}

	// EEPROM - lesen 
 	if (!ReadMem(MB_START_EEP, EEP_SIZE_SU, m_pucEepMem, piError)) {

		return FALSE;
	}

	// Groesse der Setupdaten bereits im Eep?
	ulTmp = *(ULONG*)&m_pucEepMem[P_FLASH_SIZE_SU];
	ChangeOrderLong((unsigned char*)&ulTmp);
	// Liegt im moeglichen Bereich?
	if (!(ulTmp > FLASH_SIZE_SU) && !(ulTmp == 0)) {

		ulSize = ulTmp;
	}

	// Flashbereich - lesen
	if (!ReadMem(FLASH_KONFIG_START, ulSize, m_pucFlashMem, piError)) {

		return FALSE;
	}

	// Konfiguration - lesen
	if (!ReadMem(MB_START_CONFIG, CONFIG_SIZE_SU, m_pucConfigMem, piError)) {

		return FALSE;
	}

	int i;
	TAE3Extended extBegin;
	// Erweiterter Flash vorhanden?
	if ((m_usLastDeviceVersion == 0x0400) || 
		(m_usLastDeviceVersion == 0x0401) ||
		(m_usLastDeviceVersion == 0x0402) ||
		(m_usLastDeviceVersion == 0x0500) ||
		(m_usLastDeviceVersion == 0x0501) ||
		(m_usLastDeviceVersion == 0x0502) ) {

		// Schreiben der erweiterten Struktur.
		memset(m_pucExtendedMem, 0, EXTENDED_SIZE_SU);

		memset(&extBegin, 0, sizeof(TAE3Extended));

		extBegin.ulCheck = EXTENDED_CHECK;
		extBegin.ulSize  = sizeof(TAE3Extended); 
		extBegin.pulIOModule = sizeof(TAE3Extended);
		extBegin.pulIODigital = extBegin.pulIOModule + sizeof(TAE3IOModule);
		extBegin.pulMDEDigital = extBegin.pulIODigital + 8 * sizeof(TAE3DigitalMDE);
		extBegin.pulEKArray = extBegin.pulMDEDigital + 4 * sizeof(TAE3DigitalMDE);
		extBegin.pulFreeScope = 0;  // DLL setzt nach auslesen auf 0
		extBegin.pulEntranceTables = 0;
		extBegin.pulEKArrayEntrance = 0;
		extBegin.pulValueGVx = 0;
		// Flag fuer gelesene Setupdaten setzen
		extBegin.ulFlags = EXTF_DOWNLOAD;

		for (i=0; i < sizeof(TAE3Extended); i++) {
		
			m_pucExtendedMem[i] = ((UCHAR*)&extBegin)[i];
		}
	}

	// Ab Setup - Version 3
	if ((m_usLastDeviceVersion == 0x0400) || 
		(m_usLastDeviceVersion == 0x0401) ||
		(m_usLastDeviceVersion == 0x0402) ||
		(m_usLastDeviceVersion == 0x0500) ||
		(m_usLastDeviceVersion == 0x0501) ||
		(m_usLastDeviceVersion == 0x0502) ) {

		// IOModul - Konfiguration - lesen
		if (!ReadMem(MB_START_IOMODULE, sizeof(TAE3IOModule), &m_pucExtendedMem[extBegin.pulIOModule], piError)) {

			return FALSE;
		}

		// IOModul - DigitalMDE - lesen
		for (i=0; i < 8; i++) {
			if (!ReadMem(MB_DIGITAL_IO + i * sizeof(TAE3DigitalMDE), sizeof(TAE3DigitalMDE), 
				&m_pucExtendedMem[extBegin.pulIODigital + i * sizeof(TAE3DigitalMDE)], piError)) {

				return FALSE;
			}
		}

		// Interne MDE - DigitalMDE - lesen
		for (i=0; i < 4; i++) {
			if (!ReadMem(MB_DIGITAL_MDE + i * sizeof(TAE3DigitalMDE), sizeof(TAE3DigitalMDE), 
				&m_pucExtendedMem[extBegin.pulMDEDigital + i * sizeof(TAE3DigitalMDE)], piError)) {

				return FALSE;
			}
		}

		// Groesse der MDE - Eingabeketten bereits im Eep?
		ulTmp = *(ULONG*)&m_pucEepMem[P_MDE_EK_SIZE];
		ChangeOrderLong((unsigned char*)&ulTmp);
		// Liegt im moeglichen Bereich?
		if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0)) {

			ulSize = ulTmp;
		} else {
		
			ulSize = EXTENDED_SIZE_SU - extBegin.pulEKArray;
		}

		// MDE - Eingabeketten - lesen
		// Bei 4.0 kommt erst 1Byte Anzahl im Array
		// Ab 4.1 kommen erst 4Byte Anzahl im Array
		if (!ReadMem(MB_MDE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArray], piError)) {

			return FALSE;
		}

		if ((m_usLastDeviceVersion == 0x0401) ||
			(m_usLastDeviceVersion == 0x0402) ||
			(m_usLastDeviceVersion == 0x0500) ||
			(m_usLastDeviceVersion == 0x0501) ||
			(m_usLastDeviceVersion == 0x0502) ) {

			// Dieses sollte nicht sein!, fehlende MDE - Groesse.
			if (ulSize == (EXTENDED_SIZE_SU - extBegin.pulEKArray)) {

				*piError = ERROR_FORMAT;
				return FALSE;
			}

			// Da die Tabellendefinitionen fuer die Zutrittskontrolle eine
			// feste Struktur haben, wird aufgrund des 0 - Zeigers im Strukt
			// und des Flags fuer Auslesen, im Setup das Einfuegen vorgenommen.
			extBegin.pulEKArrayEntrance = extBegin.pulEKArray + ulSize;

			// Groesse der Zutrittskontrolle - Eingabeketten bereits im Eep?
			ulTmp = *(ULONG*)&m_pucEepMem[P_ENTRANCE_EK_SIZE];
			ChangeOrderLong((unsigned char*)&ulTmp);
			// Liegt im moeglichen Bereich?
			if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0)) {

				ulSize = ulTmp;
			} else {
			
				*piError = ERROR_FORMAT;
				return FALSE;
			}

			// Zutritts - Eingabeketten - lesen
			if (!ReadMem(MB_ENTRANCE_EK, ulSize, &m_pucExtendedMem[extBegin.pulEKArrayEntrance], piError)) {

				return FALSE;
			}

			// Der Zeiger auf den naechsten freien Speicherplatz in der
			// erweiterten Struktur muss aufgrund der Abwaertskompatibilitaet
			// auf 0 gesetzt bleiben.
			for (i=0; i < sizeof(TAE3Extended); i++) {
			
				m_pucExtendedMem[i] = ((UCHAR*)&extBegin)[i];
			}
		}
	}

	temp = 0; // Bedienstopp ruecksetzten
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

		return FALSE;
	}

	return TRUE;
}

int CAE3Channel::CheckDevice(unsigned char ucBusNum, int *piError, int iPollRetry)
{
	// Testen ob Schnittstelle geoeffnet
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	for (m_iDevicePollRetry=1; m_iDevicePollRetry <= iPollRetry; m_iDevicePollRetry++) {

		int tmp;
		if (ReadMem(0ul, 2, (unsigned char*)&tmp, piError)) {

			return TRUE;
		}

		Sleep(20);
	}

	// Letzter Fehler ist aktueller
	return FALSE;
}

int CAE3Channel::Restart(unsigned char ucBusNum, int *piError)
{
	int tmp = 0;

	// Testen ob Schnittstelle geoeffnet
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	if (WriteMem(MB_RESTART, 1, (unsigned char*)&tmp, piError)) {

		return TRUE;
	}

	// Letzer Fehler ist aktueller
	return FALSE;
}

int CAE3Channel::CheckComVersion(unsigned char ucBusNum, int *piError)
{
	unsigned char Buf[4];
	
	// Wechsel nur bei -nicht Uebereinstimmung-, wegen Datensatzpuffer loeschen
	if (GetBusNum() != ucBusNum) {

		if (!ChangeBusNum(ucBusNum)) {

			*piError = ERROR_BUS_NUM;
			return FALSE;
		}
	}

	// Version fuer Setup lesen
	if (!ReadMem(MB_GET_VERSION, 2, Buf, piError)) {

		ClearCom();
		return FALSE;
	}

	// Version fuer Kommunikation
	
	// Ab Firmware 04.00.23 Listen datensatzweise laden
	if ((Buf[0] == 5 && Buf[1] == 2) ||
		// Ab Firmware 04.00.21 Listen im Hintergrund laden
		(Buf[0] == 5 && Buf[1] == 1) ||
		// Ab Firmware 04.00.19 des PZEMasterIV
		(Buf[0] == 5 && Buf[1] == 0) ||
		// Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
		(Buf[0] == 4 && Buf[1] == 2) ||
		// Ab Firmware 03.00.12 mit Zutrittskontrolle
		(Buf[0] == 4 && Buf[1] == 1) ||
	    // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
	    (Buf[0] == 4 && Buf[1] == 0) ||
		// Ab Firmware 2.8 mit 16kByte Flash fuer Setupdaten
		(Buf[0] == 3 && Buf[1] == 0) ||
		// Firmware 2.6, 2.7 mit 12kByte Flash fuer Setupdaten
		(Buf[0] == 2 && Buf[1] == 1) ||
		// <= Firmware 2.5 mit 8KByte Flash fuer Setupdaten
		(Buf[0] == 2 && Buf[1] == 0) ||
		(Buf[0] == 1 && Buf[1] == 3) || 
		(Buf[0] == 1 && Buf[1] == 2)) {
	

		// Version als Rueckgabe
		return (Buf[0] << 8) | Buf[1];
	}

	*piError = ERROR_FW_VERSION;
	return FALSE;
}

int CAE3Channel::GetSerialNum(unsigned char ucBusNum, int *piSerialNum, int *piError)
{
	unsigned char buf[10];

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Seriennummer und Optionsbyte lesen
	if (!ReadMem(MB_SERIEN_MEMORY, 4, buf, piError)) {

		return FALSE;
	}

	*piSerialNum = (int)(buf[2] << 8);
	*piSerialNum |= (int)(buf[3]);

	return TRUE;
}

// ListenBuffer in AE laden
int CAE3Channel::LoadListData(unsigned char ucBusNum, int *piError)
{

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	if (m_ulListSize == (MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG))) {

		*piError = ERROR_LIST ;
		return FALSE; // keine Listen vorhanden
	}

	unsigned long listStart = m_ulListStart;
	unsigned char temp = 1; // Bedienstopp setzen
	// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
	// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
	// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
	switch(m_usLastDeviceVersion) {
	case 0x0501:
		// Groesse pruefen
		if (m_ulListSize > FLASH_LIST_START) {

			*piError = ERROR_LIST;
			return FALSE;
		}
		break;
	case 0x0502: // 512kByte Listendaten
		listStart = FLASH_LIST_START_NEW;
		break;
	/*
	case 0x0503:
		// 2004-11-26
		//@@: Listenfunktion fuer Listentest umleiten
		return LoadNewListData(ucBusNum, piError);
		break;
	*/

	default:
		// Groesse pruefen
		if (m_ulListSize > FLASH_LIST_START) {

			*piError = ERROR_LIST;
			return FALSE;
		}

		if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}
	}

	// Flash schreiben 
	if (!WriteMem(listStart, m_ulListSize, m_pucListMem, piError, 32)) {

		return FALSE;
	}

	temp = 0; // Bedienstopp ruecksetzen
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

		return FALSE;
	}

	return TRUE;
}

// Testroutine fuer neue Listentechnik
int CAE3Channel::LoadNewListData(unsigned char ucBusNum, int *piError)
{
	return FALSE;
/*
	unsigned char buf[0x100];
	int tableNum, recordNum, datPos, pos, tab = TRUE, Err;
	ULONG tablePos, tableSize, recordCount, recordSize, add, tmp;
	// Listendaten Tabellenweise uebertragen
	for (tableNum=0; *(ULONG*)&m_pucListMem[tableNum * sizeof(ULONG)]; tableNum++) {

		buf[0] = L_EINFACH; // Tabellentyp
		buf[1] = 0xFF;      // Spalte sortiert
		//sprintf((char*)&buf[2], "L%i", tableNum); // Tabellenname
		//sprintf((char*)&buf[2], "Gehtgrnde"); // Tabellenname
		if (!GetListName(ucBusNum, tableNum, (char*)&buf[2], piError)) {

			// Liste war nicht vorhanden?
			if (*piError == ERROR_NO) {

				*piError = ERROR_NO_LD;
			}

			return FALSE;
		}
		TRACE("LstNeuStart(%s, %i, %i)\n", (char*)&buf[2], buf[0], buf[1]);


		// Befehl aufbauen
		add = LB_NEW_START | LB_DATEN_OK;
		add <<= 24;

		// Flash schreiben 
		if (!WriteMem(add, 2 + strlen((char*)&buf[2]) + 1, buf, &Err)) {

			// Fehlercode umsetzen
			*piError = g_iErrorMap[Err];
			return FALSE;
		}

		// Position im Datenarray
		tablePos = *(ULONG*)&m_pucListMem[tableNum * sizeof(ULONG)];
		ChangeOrderLong((UCHAR*)&tablePos);
		tablePos -= m_ulListStart;

		// Groesse der Tabelle ermitteln
		if (*(ULONG*)&m_pucListMem[(tableNum + 1) * sizeof(ULONG)]) {

			tmp = *(ULONG*)&m_pucListMem[(tableNum + 1) * sizeof(ULONG)];
			ChangeOrderLong((UCHAR*)&tmp);
			tmp -= m_ulListStart;

			tmp -= tablePos;
			tableSize = tmp - sizeof(ULONG);

		} else {

			tmp = m_ulListSize;// - MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG);
			tmp -= tablePos;

			tableSize = tmp - sizeof(ULONG);

		}
		// Anzahl Datensaetze in Tabelle
		recordCount = *(ULONG*)&m_pucListMem[tablePos];
		ChangeOrderLong((UCHAR*)&recordCount);

		recordSize = tableSize / recordCount;

		// Befehl aufbauen
		add = LB_DATEN | LB_DATEN_OK;
		add <<= 24;

		// Datensaetze umwandeln und uebertragen
		for (recordNum=0; recordNum < (int)recordCount; recordNum++) {

			// Datensatzformat von fester Feldlaenge zu: durch Tab getrennte Felder umwandeln
			for (datPos=0, pos=0; pos < (int)recordSize; pos++) {

				if (m_pucListMem[tablePos + sizeof(ULONG) + recordNum * recordSize + pos]) {

					buf[datPos++] = m_pucListMem[tablePos + sizeof(ULONG) + recordNum * recordSize + pos];
					tab = TRUE;

				} else if (tab) { // Erstes NULL wird zu Tab und folgende werden ignoriert

					buf[datPos++] = (unsigned char)'\t';
					tab = FALSE;
				}
			}

			buf[datPos - 1] = 0; // Letzten Tab durch NULL ersetzen.
			TRACE("LstNeuDaten [%s]\n", (char*)buf);

			// Flash schreiben 
			if (!WriteMem(add, strlen((char*)buf) + 1, buf, &Err)) {

				// Fehlercode umsetzen
				*piError = g_iErrorMap[Err];
				return FALSE;
			}

		}

		TRACE("LstNeuEnd %i\n", tableNum);

		// NULL als Daten versenden
		buf[0] = 0;

		// Befehl aufbauen
		add = LB_NEW_END | LB_DATEN_OK;
		add <<= 24;

		// Flash schreiben 
		if (!WriteMem(add, 1, buf, &Err)) {

			// Fehlercode umsetzen
			*piError = g_iErrorMap[Err];
			return FALSE;
		}

	}

	return TRUE;
*/
}


// ListenBuffer in AE laden
int CAE3Channel::LoadEntranceListData(unsigned char ucBusNum, int iListNum, int *piError)
{

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Alle Listen laden?
	if (iListNum == -1) {

		for (int i=0; i < 4; i++) {
			if (m_ulEntranceListSize[i] == 0) {

				*piError = ERROR_LIST ;
				return FALSE; // keine Listen vorhanden
			}
		}
	} else {

		if (m_ulEntranceListSize[iListNum] == 0) {

			*piError = ERROR_LIST ;
			return FALSE; // keine Listen vorhanden
		}
	}

	unsigned char temp = 1; // Bedienstopp setzen
	// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
	// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
	// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
	switch(m_usLastDeviceVersion) {
	case 0x0501:
	case 0x0502:
		break;

	default:
		if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

			return FALSE;
		}
	}

	switch(iListNum) {
	case -1:
	case 0:
		// Flash schreiben 
		if (!WriteMem(MB_ENTRANCE_TABLE_0, m_ulCompressedListSize[0], m_pucCompressedListMem[0], piError, 32)) {

			return FALSE;
		}
		if (iListNum != -1) break;
	case 1:
		// Flash schreiben 
		if (!WriteMem(MB_ENTRANCE_TABLE_1, m_ulCompressedListSize[1], m_pucCompressedListMem[1], piError, 32)) {

			return FALSE;
		}
		if (iListNum != -1) break;
	case 2:
		// Flash schreiben 
		if (!WriteMem(MB_ENTRANCE_TABLE_2, m_ulCompressedListSize[2], m_pucCompressedListMem[2], piError, 32)) {

			return FALSE;
		}
		if (iListNum != -1) break;
	case 3:
		// Flash schreiben 
		if (!WriteMem(MB_ENTRANCE_TABLE_3, m_ulCompressedListSize[3], m_pucCompressedListMem[3], piError, 32)) {

			return FALSE;
		}
	}

	temp = 0; // Bedienstopp ruecksetzen
	if (!WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {

		return FALSE;
	}

	return TRUE;
}

// ########################################################
// Alte Datensatzfunktionalitaet
// ########################################################
// Veraltete Methoden
#ifdef _WIN32
// Lesen aller Daten in den internen Speicher
// Rueckgabewerte:	1 - Lesen durchgefuehrt
//					0 - Lesen gescheitert
int CAE3Channel::ReadDatas(unsigned char ucBusNum, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int ret = -1;

	// Senden des Lese - Befehls
	m_pucTxBuf[0] = ucBusNum;				    // Bus-Nummer
	m_pucTxBuf[1] = 0x00; m_pucTxBuf[2] = 0x06;	// Befehlslaenge = 6
	m_pucTxBuf[3] = CMD_READ;				    // Befehl: Lesen
	MkMBPaket(m_pucTxBuf, 4);

	if(!Command(m_pucTxBuf, 6, pucRxBuf, &ret, piError)) {
		switch(*piError) {
		case ERROR_NO_ANSWER:
		case ERROR_AT_SENDING:
			
			return FALSE;
			break;
		case ERROR_SENDER:
		case ERROR_AT_RECIEVING:
			if(!ReReadDatas(ucBusNum, pucRxBuf, piRxCount, piError)) {
				
				return FALSE;
			}
		}
	}

	if(ret < 6) { // Paket kleiner als Mindestlaenge?

		// Daten wiederholt anfordern mit ReReadDatas
		if(!ReReadDatas(ucBusNum, pucRxBuf, piRxCount, piError)) {

			return FALSE;
		}
	}
	
	if(!CheckMBPaket(pucRxBuf, ret)) { // Pruefsumme OK?
		// Daten wiederholt anfordern mit ReReadDatas
		if(!ReReadDatas(ucBusNum, pucRxBuf, piRxCount, piError)){

			return FALSE;
		}
	}
	
	*piRxCount = ret;

	return TRUE;
}
#endif

#ifdef _WIN32
int CAE3Channel::ReReadDatas(unsigned char ucBusNum, unsigned char *pucRxBuf, int *piRxCount, int *piError)
{
	int ret = -1;
	
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	// Empfangsbuffer loeschen
	ClearCom();
	
	// Senden des Lese - Befehls
	m_pucTxBuf[0] = ucBusNum;				    // Bus-Nummer
	m_pucTxBuf[1] = 0x00; m_pucTxBuf[2] = 0x06;	// Befehlslaenge = 6
	m_pucTxBuf[3] = CMD_READ_AGAIN;	            // Befehl: Lesen
	MkMBPaket(m_pucTxBuf, 4);
	
	if(!Command(m_pucTxBuf, 6, pucRxBuf, &ret, piError)) {

		return FALSE;
	}
	
	if(ret < 6) { // Paket kleiner als Mindestlaenge?

		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	if(!CheckMBPaket(pucRxBuf, ret)){ // Pruefsumme OK?

		*piError = ERROR_CRC;
		return FALSE;
	}
	
	*piRxCount = ret;
	return TRUE;
}
#endif

#ifdef _WIN32
// Liest alle Daten des Geraets mit ucBusNum in den internen Buffer der DLL
// Rueckgabewerte:
// -1 - Fehler
//  0 - Keine Daten vorhanden
//  1 - Offline Daten vorhanden
//  2 - Online Daten vorhanden
int CAE3Channel::OldCollectData(unsigned char ucBusNum, int *piError)
{
	int retBytes,
		localError = 0, 
		AnzPakete = 0, 
		LBCnt = 0, 
		Gleich, 
		Ret, 
		Err;

	long Size = 0, 
		counter;

	bool DataAdded = false;

	unsigned char 
		Buffer[256 + 30], 
		LastBuffer[256 + 30];

	unsigned char *NutzdatenBuffer, *TempBuffer;
	NutzdatenBuffer = NULL;
	TempBuffer = NULL;

	while (1) {

		if(!ReadDatas(ucBusNum, Buffer, &retBytes, &localError)) {

			if (LBCnt > 0) {

				// letztes Paket nochmals lesen 
				Err = FALSE;
				counter = 0;
				
				do {
					if (counter) {
						// zur Sicherheit Buffer leeren
						ClearCom();
					}

					Ret = ReReadDatas(ucBusNum, Buffer, &retBytes, &localError);

				} while(counter++ < 3 && !Ret);

				if (!Ret) {
					
					Err = TRUE;
				} else {

					// Pruefen ob Paket mit letztem gueltigen Paket identisch
					Gleich = TRUE;
					if (LBCnt != retBytes-6) {
						
						Err = TRUE;	// ungleiche Anzahl Datenbyte
					} else {
						
						for (counter=0; counter < LBCnt; counter++) {

							if (LastBuffer[counter] != Buffer[counter+4]) {
								
								Gleich = FALSE;
							}
						}
					}
				}

				if (!Err && Gleich) {
					// letztes gueltige und aktuell gelesenes Paket sind identisch
					// -> naechstes Paket lesen
					counter = 0;
					do {

						Ret = ReadDatas(ucBusNum, Buffer, &retBytes, &localError);

					} while(counter++ < 3 && !Ret);

					if (!Ret) {
						
						Err = TRUE;
					}
				}
				//	Ret = ReReadDatas(ucBusNum,Buffer,retBytes,localError);

				if (Err) {

					*piError = localError;
					return -1;
				}

			} else {

				*piError = localError;
				return -1;
			}
		}

		//Sind neue Daten eingelesen worden
		if ((Buffer[3] == ANSW_NODAT) || (retBytes == 6) || (retBytes == -1)) {

			break;
		}

		AnzPakete++;
		
		// Letztes gueltige Paket sichern (fuer Vergleich bei Kommunikationsfehlern)
		if ((Buffer[3] == ANSW_DAT_NOANSW || Buffer[3] == ANSW_DAT_ANSW) && retBytes > 6) {

			LBCnt = retBytes - 6;
			for (counter=0; counter < LBCnt; counter++) {
				
				LastBuffer[counter] = Buffer[counter + 4];
			}
		}

		//Es wird keine Antwort vom PC erwartet
		if ((Buffer[3] == ANSW_DAT_NOANSW || Buffer[3] == CMD_READ_AGAIN)  && retBytes > 6) {

			m_DataList.SetState(Buffer[3]);

			if (Size != 0) {

				if (TempBuffer) {
					
					delete[] TempBuffer;
				}

				TempBuffer = new unsigned char [Size];
				
				for (counter=0; counter < Size; counter++){

					TempBuffer[counter] = NutzdatenBuffer[counter];
				}
			}

			if (NutzdatenBuffer) {
				
				delete[] NutzdatenBuffer;
			}

			NutzdatenBuffer = new unsigned char [Size+retBytes-6];
			
			for (counter=0; counter < Size; counter++) {

				NutzdatenBuffer[counter] = TempBuffer[counter];
			}

			// Neue Daten anhaengen
			for (counter=0; counter < retBytes - 6; counter++) {

				NutzdatenBuffer[counter + Size] = Buffer[counter + 4];
			}

			Size += (retBytes - 6);

			DataAdded = true;
		}

		// Es wird eine Antwort vom PC erwartet
		if ((Buffer[3] == ANSW_DAT_ANSW) && retBytes > 6) {

			m_DataList.SetState(Buffer[3]);
			
			if (Size != 0) {

				if (TempBuffer) {
					
					delete[] TempBuffer;
				}

				TempBuffer = new unsigned char[Size];
				
				for (counter=0; counter < Size; counter++) {

					TempBuffer[counter] = NutzdatenBuffer[counter];
				}
			}

			if (NutzdatenBuffer) {
				
				delete[] NutzdatenBuffer;
			}

			NutzdatenBuffer = new unsigned char[Size + retBytes - 6];

			for (counter=0; counter < Size; counter++) {

				NutzdatenBuffer[counter] = TempBuffer[counter];
			}

			// Neue Daten anhaengen
			for (counter=0; counter < retBytes - 6; counter++) {

				NutzdatenBuffer[counter + Size] = Buffer[counter + 4];
			}

			Size += (retBytes - 6);
			
			DataAdded = true;
			break;
		}
	}
	//for(int i = 0; i<65536;i++) Buffer[i]=0xFF;

	if(!DataAdded) {
		
		return 0;
	}

	// Zerlegen der gesammelten Daten in Datensaetze
	AppendData(NutzdatenBuffer, Size);

	if (TempBuffer) {
		
		delete[] TempBuffer;
	}

	if (NutzdatenBuffer) {
		
		delete[] NutzdatenBuffer;
	}

	if(m_DataList.GetState() == ANSW_DAT_ANSW) {
		
		return 2;
	}

	if(m_DataList.GetState() == ANSW_DAT_NOANSW) {
		
		return 1;
	}

	return -1;
}
#endif

#ifdef _WIN32
// Rueckgabewert: 0 keine Daten vorhanden
//                1 Daten vorhanden
//                2 letzter Datensatz
// length: Anzahl der im Buffer abgespeicherten Daten
int CAE3Channel::GetDatensatz(unsigned char *pucBuffer, int *piLength, int *piError)
{
	int iLength = 0;

	m_DataList.GetData(pucBuffer, iLength);
	*piLength = iLength;

	if(iLength == 0) {

		return 0;
	} else if (m_DataList.GetSize() == 0) {

		return 2;
	} else {

		return 1;
	}
}
#endif

// ##################################################################

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int CAE3Channel::CheckData(unsigned char ucBusNum, int *piError)
{
	unsigned char ucIsData = FALSE;

	// Testen ob Schnittstelle geoeffnet
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return -1;
	}

	// Busnummer wechsel, ohne Daten zu loeschen.
	if (!CAE3Modbus::ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return -1;
	}

	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return -1;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		// FW Ok.
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return -1;
	}

	// Abfrage ob Daten vorhanden
	if (!ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {
		
		return -1;
	}

	return ucIsData;
}

// Schliesst Relais mit Nummer fuer Zeit.
int CAE3Channel::CloseRelay(unsigned char ucBusNum, int iNum, int iTime, int *piError)
{
	if (m_ChannelType == Undef){

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int iResult = -1, iError;

	// Senden des Lese - Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		m_pucTxBuf[3] = MB_WRITE;	// Befehl: Relais schreiben.
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Adresse fuer Relais
	CAE3Modbus::ULongToBinMSBFirst(&m_pucTxBuf[4], MB_RELAY);

	m_pucTxBuf[8] = 0x80;           // Relais schliessen
	m_pucTxBuf[8] |= (iNum & 0x0F); // Relaisnummer

	m_pucTxBuf[9] = 0;

	// Zeit die Relais geschlossen wird.
	UIntToBinMSBFirst(&m_pucTxBuf[10], iTime);
	
	MkMBPaket(m_pucTxBuf, 12);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, &iError)){
			
			*piError = iError;
			Sleep(50);
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != 14){
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// CRC - pruefen
	if (!CheckMBPaket(m_pucRxBuf, iResult)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	return TRUE;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int CAE3Channel::GetRelayState(unsigned char ucBusNum, int iNum, int *piState, int *piTime, int *piError)
{
	if (m_ChannelType == Undef){

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int iResult = FALSE;

	// Senden des Lesen-Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus - Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		m_pucTxBuf[3] = MB_READ;	// Befehl: Relais lesen
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Adresse fuer Relais
	CAE3Modbus::ULongToBinMSBFirst(&m_pucTxBuf[4], MB_RELAY);

	m_pucTxBuf[8] = (iNum & 0x0F); // Nummer des Relais
	m_pucTxBuf[9] = 0;
	// Dauer
	m_pucTxBuf[10] = 0;
	m_pucTxBuf[11] = 0;

	MkMBPaket(m_pucTxBuf, 12);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, piError)){
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != 14){
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// CRC - pruefen
	if (!CheckMBPaket(m_pucRxBuf, iResult)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	if (m_pucRxBuf[8] & 0x80) {
	
		*piState = 1;
		*piTime = BinToUIntMSBFirst(&m_pucRxBuf[10]);
	} else {
	
		*piState = 0; // nicht gesetzt
		*piTime = 0;  // Zeit nicht relevant
	}

	return TRUE;
}

// Oeffnet Relais mit Nummer.
int CAE3Channel::OpenRelay(unsigned char ucBusNum, int iNum, int *piError)
{
	if (m_ChannelType == Undef){

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int iResult = FALSE;

	// Senden des Lese - Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x0E;		// Befehlslaenge = 14
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		m_pucTxBuf[3] = MB_WRITE;	// Befehl: Relais schreiben
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Adresse fuer Relais
	CAE3Modbus::ULongToBinMSBFirst(&m_pucTxBuf[4], MB_RELAY);

	m_pucTxBuf[8] = (iNum & 0x0F);
	m_pucTxBuf[9] = 0x00;
	m_pucTxBuf[10] = 0x00;
	m_pucTxBuf[11] = 0x00;

	MkMBPaket(m_pucTxBuf, 12);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, 14, m_pucRxBuf, &iResult, piError)){
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != 14){
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// CRC - pruefen
	if (!CheckMBPaket(m_pucRxBuf, iResult)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	return TRUE;
}

// Daten einer Variablen setzen.
int CAE3Channel::SetGlobVar(unsigned char ucBusNum, unsigned char *pucVar, int iType, unsigned char *pucValue, int *piError)
{
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	int iResult = FALSE;

	// Laenge ermitteln.
	int iValueLength = strlen((const char*)pucValue);

	// Senden des Lesen-Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus-Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x0C;		// Rahmen Befehlslaenge = 12
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0300:
		// Maximal 20 Zeichen schreiben.
		if (iValueLength > 20) {
			
			*piError = ERROR_RANGE;
			return FALSE;
		}
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		// Maximal 60 Zeichen schreiben.
		if (iValueLength > 60) {
			
			*piError = ERROR_RANGE;
			return FALSE;
		}
		iValueLength++; // 0 terminie

		m_pucTxBuf[3] = MB_WRITE;	// Befehl: Globale Variable schreiben.
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Adresse fuer Globale Variable
	CAE3Modbus::ULongToBinMSBFirst(&m_pucTxBuf[4], MB_GLOBAL_VAR);

	unsigned int iNameLength = 0;
	switch(iType) {
	case 0:
		m_pucTxBuf[8] = (*pucVar & 0x7F); // Nummer der zu lesenden Globalen Variable
		break;
	case 1:
		m_pucTxBuf[8] = 0x80;
		iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
		strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
		m_pucTxBuf[10 + (iNameLength - 1)] = 0;
		break;
	case 2:
		m_pucTxBuf[8] = 0x80;
		iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
		strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
		m_pucTxBuf[10 + (iNameLength - 1)] = 0;
		break;
	default:
		*piError = ERROR_RANGE;
		return FALSE;
	}

	m_pucTxBuf[9] = 0;

	// Wert in Paket kopieren
	strcpy((char*)&m_pucTxBuf[10 + iNameLength], (const char*)pucValue);
	m_pucTxBuf[10 + iNameLength + (iValueLength - 1)] = 0;

	UIntToBinMSBFirst(&m_pucTxBuf[1], m_pucTxBuf[2] + iNameLength + iValueLength + 2);
	MkMBPaket(m_pucTxBuf, BinToUIntMSBFirst(&m_pucTxBuf[1]) - 2);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, BinToUIntMSBFirst(&m_pucTxBuf[1]), m_pucRxBuf, &iResult, piError)){
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != BinToUIntMSBFirst(&m_pucRxBuf[1])) {
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	if(BinToUIntMSBFirst(&m_pucRxBuf[1]) == 6) {
		
		// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
		// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
		if (m_pucRxBuf[3] & 0x80) {
			// Fehler liegt bei Wert
			*piError = ERROR_VAR_VALUE;
		} else {
			// Variable nicht bekannt
			*piError = ERROR_NOT_DEFINED;
		}

		return FALSE;
	}

	// CRC - pruefen
	if (!CheckMBPaket(m_pucRxBuf, iResult)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	return TRUE;
}

// Daten einer Globalen Variablen lesen.
int CAE3Channel::GetGlobVar(unsigned char ucBusNum, unsigned char *pucVar, int iType, unsigned char *pucValue, int iValueLength, int *piError)
{
	if (m_ChannelType == Undef){

		*piError = ERROR_COM_CLOSED;
		return FALSE;
	}

	// muss mindestens ein Zeichen anfordern.
	if ((iValueLength - 1) <= 0) {
	
		*piError = ERROR_RANGE;
		return FALSE;
	} 

	int iResult = FALSE;

	// Senden des Lesen-Befehls
	m_pucTxBuf[0] = ucBusNum;	// Bus - Nummer
	m_pucTxBuf[1] = 0x00; 
	m_pucTxBuf[2] = 0x0C;		// Rahmen Befehlslaenge = 12
	
	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0300:
	case 0x0400:
	case 0x0401:
	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		m_pucTxBuf[3] = MB_READ;	// Befehl: Globale Variable lesen.
		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	// Adresse fuer Globale Variable
	CAE3Modbus::ULongToBinMSBFirst(&m_pucTxBuf[4], MB_GLOBAL_VAR);

	unsigned int iNameLength = 0;
	switch(iType) {
	case 0:
		m_pucTxBuf[8] = (*pucVar & 0x7F); // Nummer der zu lesenden Globalen Variable
		break;
	case 1:
		m_pucTxBuf[8] = 0x80;
		iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
		strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
		m_pucTxBuf[10 + (iNameLength - 1)] = 0;
		break;
	case 2:
		m_pucTxBuf[8] = 0x80;
		iNameLength = (strlen((const char*)pucVar) & 0x7F) + 1;
		strcpy((char*)&m_pucTxBuf[10], (const char*)pucVar);
		m_pucTxBuf[10 + (iNameLength - 1)] = 0;
		break;
	default:
		*piError = ERROR_RANGE;
		return FALSE;
	}

	// Laenge des zur Verfuegung stehenden Puffer - Nullterminie
	m_pucTxBuf[9] = (iValueLength > 60) ? 59 : (iValueLength - 1); 

	UIntToBinMSBFirst(&m_pucTxBuf[1], m_pucTxBuf[2] + iNameLength + 2);
	MkMBPaket(m_pucTxBuf, BinToUIntMSBFirst(&m_pucTxBuf[1]) - 2);

	bool bEnd = false, bCRCFail = false;

crc_fail:

	for (int i=0; (i < 4) && !bEnd; i++) {

		if(!Command(m_pucTxBuf, BinToUIntMSBFirst(&m_pucTxBuf[1]), m_pucRxBuf, &iResult, piError)){
			
			Sleep(50);
			// Nachfolgende Zeichen verwerfen
			ClearCom();
			
		} else {
		
			bEnd = true;
		}
	}

	if (!bEnd) {
	
		return FALSE;
	}

	if(iResult != BinToUIntMSBFirst(&m_pucRxBuf[1])){
		
		// Paket ungleich der Vorgeschriebenen Antwortlaenge?
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	if(BinToUIntMSBFirst(&m_pucRxBuf[1]) == 6) {
		
		// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
		// und wenn der Wert nicht gelesen werden konnte wird das hoechste bit gesetzt.
		if (m_pucRxBuf[3] & 0x80) {
			// Fehler liegt bei Wert
			*piError = ERROR_VAR_VALUE;
		} else {
			// Variable nicht bekannt
			*piError = ERROR_NOT_DEFINED;
		}

		return FALSE;
	}

	// CRC - pruefen
	if (!CheckMBPaket(m_pucRxBuf, iResult)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			// Nachfolgende Zeichen verwerfen
			ClearCom();

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}


	// Pruefe die zu kopierende Laenge.
	if (m_pucRxBuf[9] > (iValueLength - 1)) {
	
		strncpy((char*)pucValue, (const char*)&m_pucRxBuf[10 + iNameLength], (iValueLength - 1));
		pucValue[iValueLength - 1] = 0;

		*piError = ERROR_RANGE;
		return FALSE;

	} else {

		strncpy((char*)pucValue, (const char*)&m_pucRxBuf[10 + iNameLength], m_pucRxBuf[9]);
		pucValue[m_pucRxBuf[9]] = 0;
	}

	return TRUE;
}

int CAE3Channel::GetVersionFirmware(unsigned char ucBusNum, char *buf, int *piError)
{
	unsigned char lBuf[32];

	switch(m_usLastDeviceVersion) {
	case 0x0102:
	case 0x0103:
	case 0x0200:
	case 0x0201:
	case 0x0300:
	case 0x0400:
	case 0x0401:

		*piError = ERROR_FW_VERSION;
		return FALSE;
		break;

	case 0x0402:
	case 0x0500: // PZEMasterIV
	case 0x0501:
	case 0x0502:

		break;	

	default:
		*piError = ERROR_FW_VERSION;
		return FALSE;
	}

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Firmwareversion lesen
	if (!ReadMem(MB_VERSION, 32, lBuf, piError)) {

		return FALSE;
	}
	memcpy(buf, lBuf, 32);

	return TRUE;
}

int CAE3Channel::GetVersionFirmwareFromFile(char *fileName, char *versionFirmware, int *error)
{
	int res = 0;
	// Lokale Membervariable reseten
	m_fileDevice.Reset();

	if (!(res = m_fileDevice.GetVersionFirmware(fileName, versionFirmware))) {

		switch(m_fileDevice.GetLastErrorState()) {
		case ES_FILE_OPEN:
			*error = ERROR_OPEN;
			break;

		case ES_HEAD_CRC:
			*error = ERROR_CRC;
			break;

		case ES_VERSION_WRONG:
		case ES_FW_VERSION_LENGTH:
		case ES_FUNCTION_UNAVAILABLE:
			*error = ERROR_FUNCTION_UNAVAILABLE;
			break;
		}

		return FALSE;
	}

	return res;
}

// Prueft ob Displaydaten in Ordnung.
int CAE3Channel::CheckFileDisplay(CDFFileDisplay *pFileDisplay, int *piError)
{
	unsigned long ulFontPtr, ulGraficPtr;
	int i, iFontCount, iGraficCount;
	CDFFileDisplay fileDisplay;

	// Fehlerobjekt loeschen
	*piError = ERROR_NO;

	// EEPROM - Zeiger auf Zeichensatz lesen 
 	if (!ReadMem(MB_START_EEP + 0x0F78, sizeof(unsigned long), (unsigned char*)&ulFontPtr, piError)) {

		return FALSE;
	}
	ChangeOrderLong((unsigned char*)&ulFontPtr);
	if (ulFontPtr == 0xFFFFFFFF) { // Zeiger nicht vorhanden
	
		*piError = ERROR_NO;
		return FALSE;
	}

	// EEPROM - Zeiger auf Grafik lesen 
 	if (!ReadMem(MB_START_EEP + 0x0F7C, sizeof(unsigned long), (unsigned char*)&ulGraficPtr, piError)) {

		return FALSE;
	}
	ChangeOrderLong((unsigned char*)&ulGraficPtr);
	if (ulGraficPtr == 0xFFFFFFFF) { // Zeiger nicht vorhanden
		
		*piError = ERROR_NO;
		return FALSE;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE), fileDisplay.m_memFont, piError)) {

		return FALSE;
	}
	// Anzahl Zeichensaetze
	iFontCount = fileDisplay.m_pFontNode->cnt;
	if ((iFontCount > MAX_FONT) || (iFontCount != pFileDisplay->m_pFontNode->cnt)) {

		*piError = ERROR_NO;
		return FALSE;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE), fileDisplay.m_memGrafic, piError)) {

		return FALSE;
	}
	// Anzahl Grafiken
	iGraficCount = fileDisplay.m_pGraficNode->cnt;
	if ((iGraficCount > MAX_GRAFIC) || (iGraficCount != pFileDisplay->m_pGraficNode->cnt)) {

		*piError = ERROR_NO;
		return FALSE;
	}
	// Klassenstrukturen initialisieren
	fileDisplay.Default(iFontCount, iGraficCount);

	// Strukturen aus Flash lesen 
	if (!ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE) + iFontCount * sizeof(unsigned long) + iFontCount * sizeof(FONT), fileDisplay.m_memFont, piError)) {

		return FALSE;
	}
	// CRC von Zeichensaetzen ueberpruefen
	for (i=0; i < iFontCount; i++) {
		if (fileDisplay.m_pFont[i]->crc != pFileDisplay->m_pFont[i]->crc) {

			*piError = ERROR_NO;
			return FALSE;
		}
	}

	// Strukturen aus Flash lesen 
	if (!ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE) + iGraficCount * sizeof(unsigned long) + iGraficCount * sizeof(GRAFIC), fileDisplay.m_memGrafic, piError)) {

		return FALSE;
	}
	// CRC von Grafiken ueberpruefen
	for (i=0; i < iGraficCount; i++) {
		if (fileDisplay.m_pGrafic[i]->crc != pFileDisplay->m_pGrafic[i]->crc) {

			*piError = ERROR_NO;
			return FALSE;
		}
	}

	// Zeichensatz- und Grafikdaten vorhanden und identisch
	return TRUE;
}

int CAE3Channel::UploadFileDevice(unsigned char ucBusNum, int *piError)
{
	unsigned long version;

	if (!ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// EEPROM - Bootversion lesen 
 	if (!ReadMem(MB_START_EEP + EEPROM_BOOT_VERSION, sizeof(unsigned long), (unsigned char*)&version, piError)) {

		return FALSE;
	}
/*
	if ((version != 0x00020000) && (version != 0x00020002)) {

		*piError = ERROR_BOOTVERSION;
		return FALSE;
	}
*/
	unsigned char loadDisplayData = 1;
	unsigned long sizeData = m_fileDevice.GetSizeData();
	unsigned long tmp = ~0;
	unsigned long address;
	unsigned long firstAddress = m_fileDevice.GetController()->GetProgramFirstAddress();
	unsigned long lastAddress  = m_fileDevice.GetController()->GetProgramLastAddress();
	unsigned long sizeImage   = lastAddress - firstAddress + 1;

	unsigned long crc = 0;

	// EEPROM - Setupzeiger -pKonf- loeschen
	//@@: Uebergangsbug wegen Speicherueberlappung, von Firmware und Setup, beseitigen
 	if (!WriteMem(MB_START_EEP + 4, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

		return FALSE;
	}

	// Pruefen ob die Zeichensatz und Grafikdaten vorhanden
	if (!CheckFileDisplay(m_fileDevice.GetFileDisplay(), piError)) {

		if (*piError != ERROR_NO) {
			// Fehler waehrend Kommunikation, abbruch
			return FALSE;
		}

		// Keine Displaydaten oder ungleich!

	} else {
		// Anzeigedaten vorhanden und Identisch, also nicht uebertragen
		sizeData -= m_fileDevice.GetFileDisplay()->GetSizeData();
		loadDisplayData = 0;
	}

	// EEPROM - Bootload Beginn
 	if (!WriteMem(MB_BOOTLOAD, sizeof(unsigned long), (unsigned char*)&sizeData, piError)) {

		return FALSE;
	}

	tmp = FLASH_IMAGE_BASE + firstAddress; // Basisadresse im Flash
	// EEPROM - schreiben: Startadresse des Image im externen Flash
 	if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_SOURCE, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

		return FALSE;
	}

	// EEPROM - schreiben: Zieladresse des Image
 	if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_DEST, sizeof(unsigned long), (unsigned char*)&firstAddress, piError)) {

		return FALSE;
	}

	// EEPROM - schreiben: Laenge des Images
 	if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_LEN, sizeof(unsigned long), (unsigned char*)&sizeImage, piError)) {

		return FALSE;
	}

	// CRC berechnen
	crc = GetCRC16(m_fileDevice.GetController()->GetMemProgram(firstAddress), sizeImage);

	// EEPROM - schreiben: CRC des Images
 	if (!WriteMem(MB_START_EEP + EEPROM_IMAGE_CRC, sizeof(unsigned long), (unsigned char*)&crc, piError)) {

		return FALSE;
	}

	// Imagedaten
	for (address=firstAddress; address < (lastAddress + 1); address += MAX_MB_PAKET_DATA) {

		// FLASH - schreiben
 		if (!WriteMem(FLASH_IMAGE_BASE + address, 
			(((lastAddress + 1) - address) > MAX_MB_PAKET_DATA) ? MAX_MB_PAKET_DATA : ((lastAddress + 1) - address), 
			m_fileDevice.GetController()->GetMemProgram(address), piError)) {

			return FALSE;
		}

	} 

	// Laden der Zeichensatz- und Grafikdaten?
	if (loadDisplayData) {

		// Zeichensatz- und Grafikdaten
		CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

		// Zeichensatzdaten schreiben
		for (address=0; address < pFileDisplay->m_sizeFont; address += MAX_MB_PAKET_DATA) {

			// Flash schreiben 
			if (!WriteMem(0x000B0200 + address, 
				((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeFont) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeFont - address, 
				&pFileDisplay->m_memFont[address], piError)) {

				return FALSE;
			}
		}

		// Grafikdaten schreiben
		for (address=0; address < pFileDisplay->m_sizeGrafic; address += MAX_MB_PAKET_DATA) {

			if (!WriteMem(0x000B0200 + pFileDisplay->m_sizeFont + address, 
				((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeGrafic) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeGrafic - address,
				&pFileDisplay->m_memGrafic[address], piError)) {

				return FALSE;
			}
		}

		tmp = 0;
		ChangeOrderLong((unsigned char*)&tmp);
		// EEPROM - Zeiger auf Font setzen 
 		if (!WriteMem(MB_START_EEP + 0x0F78, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

			return FALSE;
		}

		tmp = pFileDisplay->m_sizeFont;
		ChangeOrderLong((unsigned char*)&tmp);
		// EEPROM - Zeiger auf Grafik setzen 
 		if (!WriteMem(MB_START_EEP + 0x0F7C, sizeof(unsigned long), (unsigned char*)&tmp, piError)) {

			return FALSE;
		}
	}

	// Schreiben der Geraetetexte
	/* Noch einbauen!! */

	sizeData = 0;
	// EEPROM - Bootload Ende
 	if (!WriteMem(MB_BOOTLOAD, sizeof(unsigned long), (unsigned char*)&sizeData, piError)) {

		return FALSE;
	}

	return TRUE;
}

int CAE3Channel::CheckCmd(unsigned char ucBusNum, int *piError)
{
	// Testen ob Schnittstelle geoeffnet
	if (m_ChannelType == Undef) {

		*piError = ERROR_COM_CLOSED;
		return -1;
	}

	// Busnummer wechsel, ohne Daten zu loeschen.
	if (!CAE3Modbus::ChangeBusNum(ucBusNum)) {

		*piError = ERROR_BUS_NUM;
		return -1;
	}

	unsigned char buf[32];
	int len = 32;
	memset(buf, 0, sizeof(buf));
	//return m_Com.Cmd(ucBusNum, buf, &len, piError);
	return m_Com.WriteMemPaket(ucBusNum, 0, len, buf, piError);
}

