/**
 * @file  DFFileDevice.h
 * @brief Schnittstelle fuer die Klasse CDFFileDevice.
 */
//////////////////////////////////////////////////////////////////////
#ifndef _DFFILEDEVICE_H_
#define _DFFILEDEVICE_H_

#include "typedef.h"
#include "ErrorState.h"
#include "ATmega128.h"
#include "IntelHex.h"
#include "DFFileDisplay.h"
#include "DFFileDeviceText.h"

// Groesse des Verschluesselungsblocks
#define DFFILEDEVICE_SIZE_CRYPT  0x200

// Fester Wert fuer Ablage, um erfolgreiche Entschluesselung sicherzustellen
#define DFFILEDEVICE_CRYPT_FIX   0xDA59EF31
// Ablagewerte fuer Einzelbyte der Cryptblockwerte
#define DFFILEDEVICE_CF_0        43
#define DFFILEDEVICE_CF_1        71
#define DFFILEDEVICE_CF_2        89
#define DFFILEDEVICE_CF_3        36

#define DFFILEDEVICE_SN_0        34
#define DFFILEDEVICE_SN_1        17
#define DFFILEDEVICE_SN_2        98
#define DFFILEDEVICE_SN_3        63

/*
Halten der grundlegenden Dateiinformationen.
Versionsnummer, Quelldateinamen, Blockgroessen
 */
#define DFFILEDEVICE_SIZE_HEADER      0x200 // Maximale Groesze fuer Headerdaten
// Fuer das PZE Master IV
#define DFFILEDEVICE_VERSION_AEIV_PZE 0x00010000
// Fuer das Zutritt Master IV
#define DFFILEDEVICE_VERSION_AEIV_ZK  0x00020000
// Fuer das BDE Master IV
#define DFFILEDEVICE_VERSION_AEIV_BDE 0x00030000

#pragma pack(1)
struct SDFFileDeviceHeader
{
	UINT32 versionFile;       // Dateiversion (Major 1 -> PZE Master IV, Major 2 -> Zutritt Master IV, Major 3 -> BDE Master IV)
	UINT16 sizeData;          // Groesze der Headstruktur: sizeof(SDFFileDeviceHeader), wird benoetigt fuer die Berechnung der Position der crcHead - Variable
	INT8 timestamp[20];       // Datetime vom Build, im Format "2005-01-25 07:00:00\0"
	UINT32 crcProgram;        // 32bit CRC fuer Programm-Flash
	INT8 versionFirmware[32]; // Firmwareversion als String, im Beispielformat: "04.00.23.3b\0"
	//...
	UINT16 crcHead;    // 16bit CRC fuer Head (Immer letzter Wert in der Struktur; 
                       // Berechnungsmenge: (sizeof(SDFFileDeviceHeader) - sizeof(UINT16)), also ohne sich selbst!)
};
#pragma pack()

class CDFFileDevice : public CIntelHex
{
public:
	CDFFileDevice();
	virtual ~CDFFileDevice();

	void Reset();
	
	inline INT32 GetLastErrorState() { return m_errorState; };
	inline UINT32 GetHeaderVersionFile() { return m_header->versionFile; };
	inline UINT32 GetHeaderVersionTimestamp(char *timestamp) { return sprintf(timestamp, "%s", m_header->timestamp); };
	inline UINT32 GetHeaderVersionFirmware(char *versionFirmware) { return sprintf(versionFirmware, "%s", m_header->versionFirmware); };
	inline UINT8 IsCrypt() { return m_isCrypt; };
	inline void SetSerial(UINT32 numSerial) { m_numSerial = numSerial; };
	inline UINT32 GetSerial() { return m_numSerial; };
	inline UINT32 GetSerialCrypt() 
	{ 
		UINT32 numSerial;
		numSerial = m_crypt[DFFILEDEVICE_SN_0];
		numSerial |= (UINT32)m_crypt[DFFILEDEVICE_SN_1] << 8;
		numSerial |= (UINT32)m_crypt[DFFILEDEVICE_SN_2] << 16;
		numSerial |= (UINT32)m_crypt[DFFILEDEVICE_SN_3] << 24;
		return numSerial;
	}

	UINT32 GetVersionFile(INT8 *fileName);
	UINT32 GetVersionFirmware(INT8 *fileName, INT8 *versionFirmware); // max. 32 Zeichen Firmwareversion
	INT32 LoadFirmware(INT8 *fileName);
	INT32 Load(INT8 *fileName);
//	INT32 Save(INT8 *fileName, SDFFileDeviceHeader &header);
	INT32 Save(INT8 *fileName, INT8 *versionFirmware);
	//INT32 Build(INT8 *fileDevice, SDFFileDeviceHeader &header, INT8 *fileFirmware, INT8 *fileDisplayData, INT8 *fileDeviceText);
	INT32 Build(INT8 *fileDevice, INT8 *versionFirmware, INT8 *fileFirmware, INT8 *fileDisplayData, INT8 *fileDeviceText);

	UINT32 GetSizeData();
	inline CATmega128 *GetController() { return &m_controller; };
	inline CDFFileDisplay *GetFileDisplay() { return &m_fileDisplay; };

	// Verschluesseln
	INT32 Encrypt(UINT32 numSerial);
	// Entschluesseln
	INT32 Decrypt();

protected:
	INT32 m_errorState;
	// Der Dateiheader wird ueber einen Puffer verwaltet, somit ist eine
	// vergroesserte Struktur auch von aelteren Softwarestaenden lesbar und
	// eine Pruefung des crcHead durch info:sizeData moeglich.
	UINT8 m_headBuf[DFFILEDEVICE_SIZE_HEADER];
	SDFFileDeviceHeader *m_header;
	CATmega128 m_controller;
	CDFFileDisplay m_fileDisplay;
	CDFFileDeviceText m_fileDeviceText;

protected:
	UINT8 m_crypt[DFFILEDEVICE_SIZE_CRYPT];
	UINT8 m_isCrypt;
	UINT32 m_numSerial;
};

#endif

